import io

import pytest

from PIL import EpsImagePlugin, Image, features

from .helper import assert_image_similar, hopper, skip_unless_feature

HAS_GHOSTSCRIPT = EpsImagePlugin.has_ghostscript()

# Our two EPS test files (they are identical except for their bounding boxes)
FILE1 = "Tests/images/zero_bb.eps"
FILE2 = "Tests/images/non_zero_bb.eps"

# Due to palletization, we'll need to convert these to RGB after load
FILE1_COMPARE = "Tests/images/zero_bb.png"
FILE1_COMPARE_SCALE2 = "Tests/images/zero_bb_scale2.png"

FILE2_COMPARE = "Tests/images/non_zero_bb.png"
FILE2_COMPARE_SCALE2 = "Tests/images/non_zero_bb_scale2.png"

# EPS test files with binary preview
FILE3 = "Tests/images/binary_preview_map.eps"


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
def test_sanity():
    # Regular scale
    with Image.open(FILE1) as image1:
        image1.load()
        assert image1.mode == "RGB"
        assert image1.size == (460, 352)
        assert image1.format == "EPS"

    with Image.open(FILE2) as image2:
        image2.load()
        assert image2.mode == "RGB"
        assert image2.size == (360, 252)
        assert image2.format == "EPS"

    # Double scale
    with Image.open(FILE1) as image1_scale2:
        image1_scale2.load(scale=2)
        assert image1_scale2.mode == "RGB"
        assert image1_scale2.size == (920, 704)
        assert image1_scale2.format == "EPS"

    with Image.open(FILE2) as image2_scale2:
        image2_scale2.load(scale=2)
        assert image2_scale2.mode == "RGB"
        assert image2_scale2.size == (720, 504)
        assert image2_scale2.format == "EPS"


def test_invalid_file():
    invalid_file = "Tests/images/flower.jpg"

    with pytest.raises(SyntaxError):
        EpsImagePlugin.EpsImageFile(invalid_file)


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
def test_cmyk():
    with Image.open("Tests/images/pil_sample_cmyk.eps") as cmyk_image:

        assert cmyk_image.mode == "CMYK"
        assert cmyk_image.size == (100, 100)
        assert cmyk_image.format == "EPS"

        cmyk_image.load()
        assert cmyk_image.mode == "RGB"

        if features.check("jpg"):
            with Image.open("Tests/images/pil_sample_rgb.jpg") as target:
                assert_image_similar(cmyk_image, target, 10)


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
def test_showpage():
    # See https://github.com/python-pillow/Pillow/issues/2615
    with Image.open("Tests/images/reqd_showpage.eps") as plot_image:
        with Image.open("Tests/images/reqd_showpage.png") as target:
            # should not crash/hang
            plot_image.load()
            #  fonts could be slightly different
            assert_image_similar(plot_image, target, 6)


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
def test_file_object(tmp_path):
    # issue 479
    with Image.open(FILE1) as image1:
        with open(str(tmp_path / "temp.eps"), "wb") as fh:
            image1.save(fh, "EPS")


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
def test_iobase_object(tmp_path):
    # issue 479
    with Image.open(FILE1) as image1:
        with open(str(tmp_path / "temp_iobase.eps"), "wb") as fh:
            image1.save(fh, "EPS")


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
def test_bytesio_object():
    with open(FILE1, "rb") as f:
        img_bytes = io.BytesIO(f.read())

    with Image.open(img_bytes) as img:
        img.load()

        with Image.open(FILE1_COMPARE) as image1_scale1_compare:
            image1_scale1_compare = image1_scale1_compare.convert("RGB")
        image1_scale1_compare.load()
        assert_image_similar(img, image1_scale1_compare, 5)


def test_image_mode_not_supported(tmp_path):
    im = hopper("RGBA")
    tmpfile = str(tmp_path / "temp.eps")
    with pytest.raises(ValueError):
        im.save(tmpfile)


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
@skip_unless_feature("zlib")
def test_render_scale1():
    # We need png support for these render test

    # Zero bounding box
    with Image.open(FILE1) as image1_scale1:
        image1_scale1.load()
        with Image.open(FILE1_COMPARE) as image1_scale1_compare:
            image1_scale1_compare = image1_scale1_compare.convert("RGB")
        image1_scale1_compare.load()
        assert_image_similar(image1_scale1, image1_scale1_compare, 5)

    # Non-Zero bounding box
    with Image.open(FILE2) as image2_scale1:
        image2_scale1.load()
        with Image.open(FILE2_COMPARE) as image2_scale1_compare:
            image2_scale1_compare = image2_scale1_compare.convert("RGB")
        image2_scale1_compare.load()
        assert_image_similar(image2_scale1, image2_scale1_compare, 10)


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
@skip_unless_feature("zlib")
def test_render_scale2():
    # We need png support for these render test

    # Zero bounding box
    with Image.open(FILE1) as image1_scale2:
        image1_scale2.load(scale=2)
        with Image.open(FILE1_COMPARE_SCALE2) as image1_scale2_compare:
            image1_scale2_compare = image1_scale2_compare.convert("RGB")
        image1_scale2_compare.load()
        assert_image_similar(image1_scale2, image1_scale2_compare, 5)

    # Non-Zero bounding box
    with Image.open(FILE2) as image2_scale2:
        image2_scale2.load(scale=2)
        with Image.open(FILE2_COMPARE_SCALE2) as image2_scale2_compare:
            image2_scale2_compare = image2_scale2_compare.convert("RGB")
        image2_scale2_compare.load()
        assert_image_similar(image2_scale2, image2_scale2_compare, 10)


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
def test_resize():
    files = [FILE1, FILE2, "Tests/images/illu10_preview.eps"]
    for fn in files:
        with Image.open(fn) as im:
            new_size = (100, 100)
            im = im.resize(new_size)
            assert im.size == new_size


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
def test_thumbnail():
    # Issue #619
    # Arrange
    files = [FILE1, FILE2]
    for fn in files:
        with Image.open(FILE1) as im:
            new_size = (100, 100)
            im.thumbnail(new_size)
            assert max(im.size) == max(new_size)


def test_read_binary_preview():
    # Issue 302
    # open image with binary preview
    with Image.open(FILE3):
        pass


def test_readline(tmp_path):
    # check all the freaking line endings possible from the spec
    # test_string = u'something\r\nelse\n\rbaz\rbif\n'
    line_endings = ["\r\n", "\n", "\n\r", "\r"]
    strings = ["something", "else", "baz", "bif"]

    def _test_readline(t, ending):
        ending = "Failure with line ending: %s" % (
            "".join("%s" % ord(s) for s in ending)
        )
        assert t.readline().strip("\r\n") == "something", ending
        assert t.readline().strip("\r\n") == "else", ending
        assert t.readline().strip("\r\n") == "baz", ending
        assert t.readline().strip("\r\n") == "bif", ending

    def _test_readline_io_psfile(test_string, ending):
        f = io.BytesIO(test_string.encode("latin-1"))
        t = EpsImagePlugin.PSFile(f)
        _test_readline(t, ending)

    def _test_readline_file_psfile(test_string, ending):
        f = str(tmp_path / "temp.txt")
        with open(f, "wb") as w:
            w.write(test_string.encode("latin-1"))

        with open(f, "rb") as r:
            t = EpsImagePlugin.PSFile(r)
            _test_readline(t, ending)

    for ending in line_endings:
        s = ending.join(strings)
        _test_readline_io_psfile(s, ending)
        _test_readline_file_psfile(s, ending)


def test_open_eps():
    # https://github.com/python-pillow/Pillow/issues/1104
    # Arrange
    FILES = [
        "Tests/images/illu10_no_preview.eps",
        "Tests/images/illu10_preview.eps",
        "Tests/images/illuCS6_no_preview.eps",
        "Tests/images/illuCS6_preview.eps",
    ]

    # Act / Assert
    for filename in FILES:
        with Image.open(filename) as img:
            assert img.mode == "RGB"


@pytest.mark.skipif(not HAS_GHOSTSCRIPT, reason="Ghostscript not available")
def test_emptyline():
    # Test file includes an empty line in the header data
    emptyline_file = "Tests/images/zero_bb_emptyline.eps"

    with Image.open(emptyline_file) as image:
        image.load()
    assert image.mode == "RGB"
    assert image.size == (460, 352)
    assert image.format == "EPS"



import pytest

from PIL import DcxImagePlugin, Image

from .helper import assert_image_equal, hopper, is_pypy

# Created with ImageMagick: convert hopper.ppm hopper.dcx
TEST_FILE = "Tests/images/hopper.dcx"


def test_sanity():
    # Arrange

    # Act
    with Image.open(TEST_FILE) as im:

        # Assert
        assert im.size == (128, 128)
        assert isinstance(im, DcxImagePlugin.DcxImageFile)
        orig = hopper()
        assert_image_equal(im, orig)


@pytest.mark.skipif(is_pypy(), reason="Requires CPython")
def test_unclosed_file():
    def open():
        im = Image.open(TEST_FILE)
        im.load()

    pytest.warns(ResourceWarning, open)


def test_closed_file():
    def open():
        im = Image.open(TEST_FILE)
        im.load()
        im.close()

    pytest.warns(None, open)


def test_context_manager():
    def open():
        with Image.open(TEST_FILE) as im:
            im.load()

    pytest.warns(None, open)


def test_invalid_file():
    with open("Tests/images/flower.jpg", "rb") as fp:
        with pytest.raises(SyntaxError):
            DcxImagePlugin.DcxImageFile(fp)


def test_tell():
    # Arrange
    with Image.open(TEST_FILE) as im:

        # Act
        frame = im.tell()

        # Assert
        assert frame == 0


def test_n_frames():
    with Image.open(TEST_FILE) as im:
        assert im.n_frames == 1
        assert not im.is_animated


def test_eoferror():
    with Image.open(TEST_FILE) as im:
        n_frames = im.n_frames

        # Test seeking past the last frame
        with pytest.raises(EOFError):
            im.seek(n_frames)
        assert im.tell() < n_frames

        # Test that seeking to the last frame does not raise an error
        im.seek(n_frames - 1)


def test_seek_too_far():
    # Arrange
    with Image.open(TEST_FILE) as im:
        frame = 999  # too big on purpose

    # Act / Assert
    with pytest.raises(EOFError):
        im.seek(frame)



from .helper import hopper


def test_getcolors():
    def getcolors(mode, limit=None):
        im = hopper(mode)
        if limit:
            colors = im.getcolors(limit)
        else:
            colors = im.getcolors()
        if colors:
            return len(colors)
        return None

    assert getcolors("1") == 2
    assert getcolors("L") == 255
    assert getcolors("I") == 255
    assert getcolors("F") == 255
    assert getcolors("P") == 90  # fixed palette
    assert getcolors("RGB") is None
    assert getcolors("RGBA") is None
    assert getcolors("CMYK") is None
    assert getcolors("YCbCr") is None

    assert getcolors("L", 128) is None
    assert getcolors("L", 1024) == 255

    assert getcolors("RGB", 8192) is None
    assert getcolors("RGB", 16384) == 10100
    assert getcolors("RGB", 100000) == 10100

    assert getcolors("RGBA", 16384) == 10100
    assert getcolors("CMYK", 16384) == 10100
    assert getcolors("YCbCr", 16384) == 9329


# --------------------------------------------------------------------


def test_pack():
    # Pack problems for small tables (@PIL209)

    im = hopper().quantize(3).convert("RGB")

    expected = [
        (4039, (172, 166, 181)),
        (4385, (124, 113, 134)),
        (7960, (31, 20, 33)),
    ]

    A = im.getcolors(maxcolors=2)
    assert A is None

    A = im.getcolors(maxcolors=3)
    A.sort()
    assert A == expected

    A = im.getcolors(maxcolors=4)
    A.sort()
    assert A == expected

    A = im.getcolors(maxcolors=8)
    A.sort()
    assert A == expected

    A = im.getcolors(maxcolors=16)
    A.sort()
    assert A == expected



from PIL import ImageQt

from .helper import hopper
from .test_imageqt import PillowQPixmapTestCase

if ImageQt.qt_is_installed:
    from PIL.ImageQt import QPixmap


class TestToQPixmap(PillowQPixmapTestCase):
    def test_sanity(self, tmp_path):
        for mode in ("1", "RGB", "RGBA", "L", "P"):
            data = ImageQt.toqpixmap(hopper(mode))

            assert isinstance(data, QPixmap)
            assert not data.isNull()

            # Test saving the file
            tempfile = str(tmp_path / f"temp_{mode}.png")
            data.save(tempfile)



from PIL import Image


def test_interface():
    im = Image.new("RGBA", (1, 1), (1, 2, 3, 0))
    assert im.getpixel((0, 0)) == (1, 2, 3, 0)

    im = Image.new("RGBA", (1, 1), (1, 2, 3))
    assert im.getpixel((0, 0)) == (1, 2, 3, 255)

    im.putalpha(Image.new("L", im.size, 4))
    assert im.getpixel((0, 0)) == (1, 2, 3, 4)

    im.putalpha(5)
    assert im.getpixel((0, 0)) == (1, 2, 3, 5)


def test_promote():
    im = Image.new("L", (1, 1), 1)
    assert im.getpixel((0, 0)) == 1

    im.putalpha(2)
    assert im.mode == "LA"
    assert im.getpixel((0, 0)) == (1, 2)

    im = Image.new("P", (1, 1), 1)
    assert im.getpixel((0, 0)) == 1

    im.putalpha(2)
    assert im.mode == "PA"
    assert im.getpixel((0, 0)) == (1, 2)

    im = Image.new("RGB", (1, 1), (1, 2, 3))
    assert im.getpixel((0, 0)) == (1, 2, 3)

    im.putalpha(4)
    assert im.mode == "RGBA"
    assert im.getpixel((0, 0)) == (1, 2, 3, 4)


def test_readonly():
    im = Image.new("RGB", (1, 1), (1, 2, 3))
    im.readonly = 1

    im.putalpha(4)
    assert not im.readonly
    assert im.mode == "RGBA"
    assert im.getpixel((0, 0)) == (1, 2, 3, 4)



import copy

from PIL import Image

from .helper import hopper


def test_copy():
    croppedCoordinates = (10, 10, 20, 20)
    croppedSize = (10, 10)
    for mode in "1", "P", "L", "RGB", "I", "F":
        # Internal copy method
        im = hopper(mode)
        out = im.copy()
        assert out.mode == im.mode
        assert out.size == im.size

        # Python's copy method
        im = hopper(mode)
        out = copy.copy(im)
        assert out.mode == im.mode
        assert out.size == im.size

        # Internal copy method on a cropped image
        im = hopper(mode)
        out = im.crop(croppedCoordinates).copy()
        assert out.mode == im.mode
        assert out.size == croppedSize

        # Python's copy method on a cropped image
        im = hopper(mode)
        out = copy.copy(im.crop(croppedCoordinates))
        assert out.mode == im.mode
        assert out.size == croppedSize


def test_copy_zero():
    im = Image.new("RGB", (0, 0))
    out = im.copy()
    assert out.mode == im.mode
    assert out.size == im.size



import pytest

from PIL import CurImagePlugin, Image

TEST_FILE = "Tests/images/deerstalker.cur"


def test_sanity():
    with Image.open(TEST_FILE) as im:
        assert im.size == (32, 32)
        assert isinstance(im, CurImagePlugin.CurImageFile)
        # Check some pixel colors to ensure image is loaded properly
        assert im.getpixel((10, 1)) == (0, 0, 0, 0)
        assert im.getpixel((11, 1)) == (253, 254, 254, 1)
        assert im.getpixel((16, 16)) == (84, 87, 86, 255)


def test_invalid_file():
    invalid_file = "Tests/images/flower.jpg"

    with pytest.raises(SyntaxError):
        CurImagePlugin.CurImageFile(invalid_file)

    no_cursors_file = "Tests/images/no_cursors.cur"

    cur = CurImagePlugin.CurImageFile(TEST_FILE)
    cur.fp.close()
    with open(no_cursors_file, "rb") as cur.fp:
        with pytest.raises(TypeError):
            cur._open()



"""
Print out some handy system info like Travis CI does.

This sort of info is missing from GitHub Actions.

Requested here:
https://github.com/actions/virtual-environments/issues/79
"""
import os
import platform
import sys

print("Build system information")
print()

print("sys.version\t\t", sys.version.split("\n"))
print("os.name\t\t\t", os.name)
print("sys.platform\t\t", sys.platform)
print("platform.system()\t", platform.system())
print("platform.machine()\t", platform.machine())
print("platform.platform()\t", platform.platform())
print("platform.version()\t", platform.version())
print("platform.uname()\t", platform.uname())
if sys.platform == "darwin":
    print("platform.mac_ver()\t", platform.mac_ver())



#!/usr/bin/env python
# > pyroma .
# ------------------------------
# Checking .
# Found Pillow
# ------------------------------
# Final rating: 10/10
# Your cheese is so fresh most people think it's a cream: Mascarpone
# ------------------------------

import os
import re
import struct
import subprocess
import sys
import warnings

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext


def get_version():
    version_file = "src/PIL/_version.py"
    with open(version_file) as f:
        exec(compile(f.read(), version_file, "exec"))
    return locals()["__version__"]


NAME = "Pillow"
PILLOW_VERSION = get_version()
FREETYPE_ROOT = None
IMAGEQUANT_ROOT = None
JPEG2K_ROOT = None
JPEG_ROOT = None
LCMS_ROOT = None
TIFF_ROOT = None
ZLIB_ROOT = None


if sys.platform == "win32" and sys.version_info >= (3, 10):
    warnings.warn(
        f"Pillow {PILLOW_VERSION} does not support Python "
        f"{sys.version_info.major}.{sys.version_info.minor} and does not provide "
        "prebuilt Windows binaries. We do not recommend building from source on "
        "Windows.",
        RuntimeWarning,
    )


_IMAGING = ("decode", "encode", "map", "display", "outline", "path")

_LIB_IMAGING = (
    "Access",
    "AlphaComposite",
    "Resample",
    "Reduce",
    "Bands",
    "BcnDecode",
    "BitDecode",
    "Blend",
    "Chops",
    "ColorLUT",
    "Convert",
    "ConvertYCbCr",
    "Copy",
    "Crop",
    "Dib",
    "Draw",
    "Effects",
    "EpsEncode",
    "File",
    "Fill",
    "Filter",
    "FliDecode",
    "Geometry",
    "GetBBox",
    "GifDecode",
    "GifEncode",
    "HexDecode",
    "Histo",
    "JpegDecode",
    "JpegEncode",
    "Matrix",
    "ModeFilter",
    "Negative",
    "Offset",
    "Pack",
    "PackDecode",
    "Palette",
    "Paste",
    "Quant",
    "QuantOctree",
    "QuantHash",
    "QuantHeap",
    "PcdDecode",
    "PcxDecode",
    "PcxEncode",
    "Point",
    "RankFilter",
    "RawDecode",
    "RawEncode",
    "Storage",
    "SgiRleDecode",
    "SunRleDecode",
    "TgaRleDecode",
    "TgaRleEncode",
    "Unpack",
    "UnpackYCC",
    "UnsharpMask",
    "XbmDecode",
    "XbmEncode",
    "ZipDecode",
    "ZipEncode",
    "TiffDecode",
    "Jpeg2KDecode",
    "Jpeg2KEncode",
    "BoxBlur",
    "QuantPngQuant",
    "codec_fd",
)

DEBUG = False


class DependencyException(Exception):
    pass


class RequiredDependencyException(Exception):
    pass


PLATFORM_MINGW = os.name == "nt" and "GCC" in sys.version
PLATFORM_PYPY = hasattr(sys, "pypy_version_info")


def _dbg(s, tp=None):
    if DEBUG:
        if tp:
            print(s % tp)
            return
        print(s)


def _find_library_dirs_ldconfig():
    # Based on ctypes.util from Python 2

    if sys.platform.startswith("linux") or sys.platform.startswith("gnu"):
        if struct.calcsize("l") == 4:
            machine = os.uname()[4] + "-32"
        else:
            machine = os.uname()[4] + "-64"
        mach_map = {
            "x86_64-64": "libc6,x86-64",
            "ppc64-64": "libc6,64bit",
            "sparc64-64": "libc6,64bit",
            "s390x-64": "libc6,64bit",
            "ia64-64": "libc6,IA-64",
        }
        abi_type = mach_map.get(machine, "libc6")

        # Assuming GLIBC's ldconfig (with option -p)
        # Alpine Linux uses musl that can't print cache
        args = ["/sbin/ldconfig", "-p"]
        expr = fr".*\({abi_type}.*\) => (.*)"
        env = dict(os.environ)
        env["LC_ALL"] = "C"
        env["LANG"] = "C"

    elif sys.platform.startswith("freebsd"):
        args = ["/sbin/ldconfig", "-r"]
        expr = r".* => (.*)"
        env = {}

    try:
        p = subprocess.Popen(
            args, stderr=subprocess.DEVNULL, stdout=subprocess.PIPE, env=env
        )
    except OSError:  # E.g. command not found
        return []
    [data, _] = p.communicate()
    if isinstance(data, bytes):
        data = data.decode()

    dirs = []
    for dll in re.findall(expr, data):
        dir = os.path.dirname(dll)
        if dir not in dirs:
            dirs.append(dir)
    return dirs


def _add_directory(path, subdir, where=None):
    if subdir is None:
        return
    subdir = os.path.realpath(subdir)
    if os.path.isdir(subdir) and subdir not in path:
        if where is None:
            _dbg("Appending path %s", subdir)
            path.append(subdir)
        else:
            _dbg("Inserting path %s", subdir)
            path.insert(where, subdir)
    elif subdir in path and where is not None:
        path.remove(subdir)
        path.insert(where, subdir)


def _find_include_file(self, include):
    for directory in self.compiler.include_dirs:
        _dbg("Checking for include file %s in %s", (include, directory))
        if os.path.isfile(os.path.join(directory, include)):
            _dbg("Found %s", include)
            return 1
    return 0


def _find_library_file(self, library):
    ret = self.compiler.find_library_file(self.compiler.library_dirs, library)
    if ret:
        _dbg("Found library %s at %s", (library, ret))
    else:
        _dbg("Couldn't find library %s in %s", (library, self.compiler.library_dirs))
    return ret


def _cmd_exists(cmd):
    return any(
        os.access(os.path.join(path, cmd), os.X_OK)
        for path in os.environ["PATH"].split(os.pathsep)
    )


def _pkg_config(name):
    try:
        command = os.environ.get("PKG_CONFIG", "pkg-config")
        command_libs = [command, "--libs-only-L", name]
        command_cflags = [command, "--cflags-only-I", name]
        if not DEBUG:
            command_libs.append("--silence-errors")
            command_cflags.append("--silence-errors")
        libs = (
            subprocess.check_output(command_libs)
            .decode("utf8")
            .strip()
            .replace("-L", "")
        )
        cflags = (
            subprocess.check_output(command_cflags)
            .decode("utf8")
            .strip()
            .replace("-I", "")
        )
        return (libs, cflags)
    except Exception:
        pass


class pil_build_ext(build_ext):
    class feature:
        features = [
            "zlib",
            "jpeg",
            "tiff",
            "freetype",
            "lcms",
            "webp",
            "webpmux",
            "jpeg2000",
            "imagequant",
            "xcb",
        ]

        required = {"jpeg", "zlib"}

        def __init__(self):
            for f in self.features:
                setattr(self, f, None)

        def require(self, feat):
            return feat in self.required

        def want(self, feat):
            return getattr(self, feat) is None

        def __iter__(self):
            yield from self.features

    feature = feature()

    user_options = (
        build_ext.user_options
        + [(f"disable-{x}", None, f"Disable support for {x}") for x in feature]
        + [(f"enable-{x}", None, f"Enable support for {x}") for x in feature]
        + [
            ("disable-platform-guessing", None, "Disable platform guessing on Linux"),
            ("debug", None, "Debug logging"),
        ]
        + [("add-imaging-libs=", None, "Add libs to _imaging build")]
    )

    def initialize_options(self):
        self.disable_platform_guessing = None
        self.add_imaging_libs = ""
        build_ext.initialize_options(self)
        for x in self.feature:
            setattr(self, f"disable_{x}", None)
            setattr(self, f"enable_{x}", None)

    def finalize_options(self):
        build_ext.finalize_options(self)
        if self.debug:
            global DEBUG
            DEBUG = True
        if not self.parallel:
            # If --parallel (or -j) wasn't specified, we want to reproduce the same
            # behavior as before, that is, auto-detect the number of jobs.
            try:
                self.parallel = int(
                    os.environ.get("MAX_CONCURRENCY", min(4, os.cpu_count()))
                )
            except TypeError:
                self.parallel = None
        for x in self.feature:
            if getattr(self, f"disable_{x}"):
                setattr(self.feature, x, False)
                self.feature.required.discard(x)
                _dbg("Disabling %s", x)
                if getattr(self, f"enable_{x}"):
                    raise ValueError(
                        f"Conflicting options: --enable-{x} and --disable-{x}"
                    )
            if getattr(self, f"enable_{x}"):
                _dbg("Requiring %s", x)
                self.feature.required.add(x)

    def _update_extension(self, name, libraries, define_macros=None, include_dirs=None):
        for extension in self.extensions:
            if extension.name == name:
                extension.libraries += libraries
                if define_macros is not None:
                    extension.define_macros += define_macros
                if include_dirs is not None:
                    extension.include_dirs += include_dirs
                break

    def _remove_extension(self, name):
        for extension in self.extensions:
            if extension.name == name:
                self.extensions.remove(extension)
                break

    def build_extensions(self):

        library_dirs = []
        include_dirs = []

        pkg_config = None
        if _cmd_exists(os.environ.get("PKG_CONFIG", "pkg-config")):
            pkg_config = _pkg_config

        #
        # add configured kits
        for root_name, lib_name in dict(
            JPEG_ROOT="libjpeg",
            JPEG2K_ROOT="libopenjp2",
            TIFF_ROOT=("libtiff-5", "libtiff-4"),
            ZLIB_ROOT="zlib",
            FREETYPE_ROOT="freetype2",
            LCMS_ROOT="lcms2",
            IMAGEQUANT_ROOT="libimagequant",
        ).items():
            root = globals()[root_name]

            if root is None and root_name in os.environ:
                prefix = os.environ[root_name]
                root = (os.path.join(prefix, "lib"), os.path.join(prefix, "include"))

            if root is None and pkg_config:
                if isinstance(lib_name, tuple):
                    for lib_name2 in lib_name:
                        _dbg(f"Looking for `{lib_name2}` using pkg-config.")
                        root = pkg_config(lib_name2)
                        if root:
                            break
                else:
                    _dbg(f"Looking for `{lib_name}` using pkg-config.")
                    root = pkg_config(lib_name)

            if isinstance(root, tuple):
                lib_root, include_root = root
            else:
                lib_root = include_root = root

            _add_directory(library_dirs, lib_root)
            _add_directory(include_dirs, include_root)

        # respect CFLAGS/CPPFLAGS/LDFLAGS
        for k in ("CFLAGS", "CPPFLAGS", "LDFLAGS"):
            if k in os.environ:
                for match in re.finditer(r"-I([^\s]+)", os.environ[k]):
                    _add_directory(include_dirs, match.group(1))
                for match in re.finditer(r"-L([^\s]+)", os.environ[k]):
                    _add_directory(library_dirs, match.group(1))

        # include, rpath, if set as environment variables:
        for k in ("C_INCLUDE_PATH", "CPATH", "INCLUDE"):
            if k in os.environ:
                for d in os.environ[k].split(os.path.pathsep):
                    _add_directory(include_dirs, d)

        for k in ("LD_RUN_PATH", "LIBRARY_PATH", "LIB"):
            if k in os.environ:
                for d in os.environ[k].split(os.path.pathsep):
                    _add_directory(library_dirs, d)

        _add_directory(library_dirs, os.path.join(sys.prefix, "lib"))
        _add_directory(include_dirs, os.path.join(sys.prefix, "include"))

        #
        # add platform directories

        if self.disable_platform_guessing:
            pass

        elif sys.platform == "cygwin":
            # pythonX.Y.dll.a is in the /usr/lib/pythonX.Y/config directory
            _add_directory(
                library_dirs,
                os.path.join(
                    "/usr/lib", "python{}.{}".format(*sys.version_info), "config"
                ),
            )

        elif sys.platform == "darwin":
            # attempt to make sure we pick freetype2 over other versions
            _add_directory(include_dirs, "/sw/include/freetype2")
            _add_directory(include_dirs, "/sw/lib/freetype2/include")
            # fink installation directories
            _add_directory(library_dirs, "/sw/lib")
            _add_directory(include_dirs, "/sw/include")
            # darwin ports installation directories
            _add_directory(library_dirs, "/opt/local/lib")
            _add_directory(include_dirs, "/opt/local/include")

            # if Homebrew is installed, use its lib and include directories
            try:
                prefix = (
                    subprocess.check_output(["brew", "--prefix"])
                    .strip()
                    .decode("latin1")
                )
            except Exception:
                # Homebrew not installed
                prefix = None

            ft_prefix = None

            if prefix:
                # add Homebrew's include and lib directories
                _add_directory(library_dirs, os.path.join(prefix, "lib"))
                _add_directory(include_dirs, os.path.join(prefix, "include"))
                _add_directory(
                    include_dirs, os.path.join(prefix, "opt", "zlib", "include")
                )
                ft_prefix = os.path.join(prefix, "opt", "freetype")

            if ft_prefix and os.path.isdir(ft_prefix):
                # freetype might not be linked into Homebrew's prefix
                _add_directory(library_dirs, os.path.join(ft_prefix, "lib"))
                _add_directory(include_dirs, os.path.join(ft_prefix, "include"))
            else:
                # fall back to freetype from XQuartz if
                # Homebrew's freetype is missing
                _add_directory(library_dirs, "/usr/X11/lib")
                _add_directory(include_dirs, "/usr/X11/include")

            # SDK install path
            try:
                sdk_path = (
                    subprocess.check_output(["xcrun", "--show-sdk-path"])
                    .strip()
                    .decode("latin1")
                )
            except Exception:
                sdk_path = None
            if sdk_path:
                _add_directory(library_dirs, os.path.join(sdk_path, "usr", "lib"))
                _add_directory(include_dirs, os.path.join(sdk_path, "usr", "include"))
        elif (
            sys.platform.startswith("linux")
            or sys.platform.startswith("gnu")
            or sys.platform.startswith("freebsd")
        ):
            for dirname in _find_library_dirs_ldconfig():
                _add_directory(library_dirs, dirname)
            if sys.platform.startswith("linux") and os.environ.get(
                "ANDROID_ROOT", None
            ):
                # termux support for android.
                # system libraries (zlib) are installed in /system/lib
                # headers are at $PREFIX/include
                # user libs are at $PREFIX/lib
                _add_directory(
                    library_dirs, os.path.join(os.environ["ANDROID_ROOT"], "lib")
                )

        elif sys.platform.startswith("netbsd"):
            _add_directory(library_dirs, "/usr/pkg/lib")
            _add_directory(include_dirs, "/usr/pkg/include")

        elif sys.platform.startswith("sunos5"):
            _add_directory(library_dirs, "/opt/local/lib")
            _add_directory(include_dirs, "/opt/local/include")

        # FIXME: check /opt/stuff directories here?

        # standard locations
        if not self.disable_platform_guessing:
            _add_directory(library_dirs, "/usr/local/lib")
            _add_directory(include_dirs, "/usr/local/include")

            _add_directory(library_dirs, "/usr/lib")
            _add_directory(include_dirs, "/usr/include")
            # alpine, at least
            _add_directory(library_dirs, "/lib")

        if sys.platform == "win32":
            # on Windows, look for the OpenJPEG libraries in the location that
            # the official installer puts them
            program_files = os.environ.get("ProgramFiles", "")
            best_version = (0, 0)
            best_path = None
            for name in os.listdir(program_files):
                if name.startswith("OpenJPEG "):
                    version = tuple(int(x) for x in name[9:].strip().split("."))
                    if version > best_version:
                        best_version = version
                        best_path = os.path.join(program_files, name)

            if best_path:
                _dbg("Adding %s to search list", best_path)
                _add_directory(library_dirs, os.path.join(best_path, "lib"))
                _add_directory(include_dirs, os.path.join(best_path, "include"))

        #
        # insert new dirs *before* default libs, to avoid conflicts
        # between Python PYD stub libs and real libraries

        self.compiler.library_dirs = library_dirs + self.compiler.library_dirs
        self.compiler.include_dirs = include_dirs + self.compiler.include_dirs

        #
        # look for available libraries

        feature = self.feature

        if feature.want("zlib"):
            _dbg("Looking for zlib")
            if _find_include_file(self, "zlib.h"):
                if _find_library_file(self, "z"):
                    feature.zlib = "z"
                elif sys.platform == "win32" and _find_library_file(self, "zlib"):
                    feature.zlib = "zlib"  # alternative name

        if feature.want("jpeg"):
            _dbg("Looking for jpeg")
            if _find_include_file(self, "jpeglib.h"):
                if _find_library_file(self, "jpeg"):
                    feature.jpeg = "jpeg"
                elif sys.platform == "win32" and _find_library_file(self, "libjpeg"):
                    feature.jpeg = "libjpeg"  # alternative name

        feature.openjpeg_version = None
        if feature.want("jpeg2000"):
            _dbg("Looking for jpeg2000")
            best_version = None
            best_path = None

            # Find the best version
            for directory in self.compiler.include_dirs:
                _dbg("Checking for openjpeg-#.# in %s", directory)
                try:
                    listdir = os.listdir(directory)
                except Exception:
                    # OSError, FileNotFoundError
                    continue
                for name in listdir:
                    if name.startswith("openjpeg-") and os.path.isfile(
                        os.path.join(directory, name, "openjpeg.h")
                    ):
                        _dbg("Found openjpeg.h in %s/%s", (directory, name))
                        version = tuple(int(x) for x in name[9:].split("."))
                        if best_version is None or version > best_version:
                            best_version = version
                            best_path = os.path.join(directory, name)
                            _dbg(
                                "Best openjpeg version %s so far in %s",
                                (best_version, best_path),
                            )

            if best_version and _find_library_file(self, "openjp2"):
                # Add the directory to the include path so we can include
                # <openjpeg.h> rather than having to cope with the versioned
                # include path
                # FIXME (melvyn-sopacua):
                # At this point it's possible that best_path is already in
                # self.compiler.include_dirs. Should investigate how that is
                # possible.
                _add_directory(self.compiler.include_dirs, best_path, 0)
                feature.jpeg2000 = "openjp2"
                feature.openjpeg_version = ".".join(str(x) for x in best_version)

        if feature.want("imagequant"):
            _dbg("Looking for imagequant")
            if _find_include_file(self, "libimagequant.h"):
                if _find_library_file(self, "imagequant"):
                    feature.imagequant = "imagequant"
                elif _find_library_file(self, "libimagequant"):
                    feature.imagequant = "libimagequant"

        if feature.want("tiff"):
            _dbg("Looking for tiff")
            if _find_include_file(self, "tiff.h"):
                if _find_library_file(self, "tiff"):
                    feature.tiff = "tiff"
                if sys.platform in ["win32", "darwin"] and _find_library_file(
                    self, "libtiff"
                ):
                    feature.tiff = "libtiff"

        if feature.want("freetype"):
            _dbg("Looking for freetype")
            if _find_library_file(self, "freetype"):
                # look for freetype2 include files
                freetype_version = 0
                for subdir in self.compiler.include_dirs:
                    _dbg("Checking for include file %s in %s", ("ft2build.h", subdir))
                    if os.path.isfile(os.path.join(subdir, "ft2build.h")):
                        _dbg("Found %s in %s", ("ft2build.h", subdir))
                        freetype_version = 21
                        subdir = os.path.join(subdir, "freetype2")
                        break
                    subdir = os.path.join(subdir, "freetype2")
                    _dbg("Checking for include file %s in %s", ("ft2build.h", subdir))
                    if os.path.isfile(os.path.join(subdir, "ft2build.h")):
                        _dbg("Found %s in %s", ("ft2build.h", subdir))
                        freetype_version = 21
                        break
                if freetype_version:
                    feature.freetype = "freetype"
                    if subdir:
                        _add_directory(self.compiler.include_dirs, subdir, 0)

        if feature.want("lcms"):
            _dbg("Looking for lcms")
            if _find_include_file(self, "lcms2.h"):
                if _find_library_file(self, "lcms2"):
                    feature.lcms = "lcms2"
                elif _find_library_file(self, "lcms2_static"):
                    # alternate Windows name.
                    feature.lcms = "lcms2_static"

        if feature.want("webp"):
            _dbg("Looking for webp")
            if _find_include_file(self, "webp/encode.h") and _find_include_file(
                self, "webp/decode.h"
            ):
                # In Google's precompiled zip it is call "libwebp":
                if _find_library_file(self, "webp"):
                    feature.webp = "webp"
                elif _find_library_file(self, "libwebp"):
                    feature.webp = "libwebp"

        if feature.want("webpmux"):
            _dbg("Looking for webpmux")
            if _find_include_file(self, "webp/mux.h") and _find_include_file(
                self, "webp/demux.h"
            ):
                if _find_library_file(self, "webpmux") and _find_library_file(
                    self, "webpdemux"
                ):
                    feature.webpmux = "webpmux"
                if _find_library_file(self, "libwebpmux") and _find_library_file(
                    self, "libwebpdemux"
                ):
                    feature.webpmux = "libwebpmux"

        if feature.want("xcb"):
            _dbg("Looking for xcb")
            if _find_include_file(self, "xcb/xcb.h"):
                if _find_library_file(self, "xcb"):
                    feature.xcb = "xcb"

        for f in feature:
            if not getattr(feature, f) and feature.require(f):
                if f in ("jpeg", "zlib"):
                    raise RequiredDependencyException(f)
                raise DependencyException(f)

        #
        # core library

        libs = self.add_imaging_libs.split()
        defs = []
        if feature.jpeg:
            libs.append(feature.jpeg)
            defs.append(("HAVE_LIBJPEG", None))
        if feature.jpeg2000:
            libs.append(feature.jpeg2000)
            defs.append(("HAVE_OPENJPEG", None))
            if sys.platform == "win32" and not PLATFORM_MINGW:
                defs.append(("OPJ_STATIC", None))
        if feature.zlib:
            libs.append(feature.zlib)
            defs.append(("HAVE_LIBZ", None))
        if feature.imagequant:
            libs.append(feature.imagequant)
            defs.append(("HAVE_LIBIMAGEQUANT", None))
        if feature.tiff:
            libs.append(feature.tiff)
            defs.append(("HAVE_LIBTIFF", None))
            # FIXME the following define should be detected automatically
            #       based on system libtiff, see #4237
            if PLATFORM_MINGW:
                defs.append(("USE_WIN32_FILEIO", None))
        if feature.xcb:
            libs.append(feature.xcb)
            defs.append(("HAVE_XCB", None))
        if sys.platform == "win32":
            libs.extend(["kernel32", "user32", "gdi32"])
        if struct.unpack("h", b"\0\1")[0] == 1:
            defs.append(("WORDS_BIGENDIAN", None))

        if (
            sys.platform == "win32"
            and sys.version_info < (3, 9)
            and not (PLATFORM_PYPY or PLATFORM_MINGW)
        ):
            defs.append(("PILLOW_VERSION", f'"\\"{PILLOW_VERSION}\\""'))
        else:
            defs.append(("PILLOW_VERSION", f'"{PILLOW_VERSION}"'))

        self._update_extension("PIL._imaging", libs, defs)

        #
        # additional libraries

        if feature.freetype:
            libs = ["freetype"]
            defs = []
            self._update_extension("PIL._imagingft", libs, defs)
        else:
            self._remove_extension("PIL._imagingft")

        if feature.lcms:
            extra = []
            if sys.platform == "win32":
                extra.extend(["user32", "gdi32"])
            self._update_extension("PIL._imagingcms", [feature.lcms] + extra)
        else:
            self._remove_extension("PIL._imagingcms")

        if feature.webp:
            libs = [feature.webp]
            defs = []

            if feature.webpmux:
                defs.append(("HAVE_WEBPMUX", None))
                libs.append(feature.webpmux)
                libs.append(feature.webpmux.replace("pmux", "pdemux"))

            self._update_extension("PIL._webp", libs, defs)
        else:
            self._remove_extension("PIL._webp")

        tk_libs = ["psapi"] if sys.platform == "win32" else []
        self._update_extension("PIL._imagingtk", tk_libs)

        build_ext.build_extensions(self)

        #
        # sanity checks

        self.summary_report(feature)

    def summary_report(self, feature):

        print("-" * 68)
        print("PIL SETUP SUMMARY")
        print("-" * 68)
        print(f"version      Pillow {PILLOW_VERSION}")
        v = sys.version.split("[")
        print(f"platform     {sys.platform} {v[0].strip()}")
        for v in v[1:]:
            print(f"             [{v.strip()}")
        print("-" * 68)

        options = [
            (feature.jpeg, "JPEG"),
            (feature.jpeg2000, "OPENJPEG (JPEG2000)", feature.openjpeg_version),
            (feature.zlib, "ZLIB (PNG/ZIP)"),
            (feature.imagequant, "LIBIMAGEQUANT"),
            (feature.tiff, "LIBTIFF"),
            (feature.freetype, "FREETYPE2"),
            (feature.lcms, "LITTLECMS2"),
            (feature.webp, "WEBP"),
            (feature.webpmux, "WEBPMUX"),
            (feature.xcb, "XCB (X protocol)"),
        ]

        all = 1
        for option in options:
            if option[0]:
                version = ""
                if len(option) >= 3 and option[2]:
                    version = f" ({option[2]})"
                print(f"--- {option[1]} support available{version}")
            else:
                print(f"*** {option[1]} support not available")
                all = 0

        print("-" * 68)

        if not all:
            print("To add a missing option, make sure you have the required")
            print("library and headers.")
            print(
                "See https://pillow.readthedocs.io/en/latest/installation."
                "html#building-from-source"
            )
            print("")

        print("To check the build, run the selftest.py script.")
        print("")


def debug_build():
    return hasattr(sys, "gettotalrefcount")


files = ["src/_imaging.c"]
for src_file in _IMAGING:
    files.append("src/" + src_file + ".c")
for src_file in _LIB_IMAGING:
    files.append(os.path.join("src/libImaging", src_file + ".c"))
ext_modules = [
    Extension("PIL._imaging", files),
    Extension("PIL._imagingft", ["src/_imagingft.c"]),
    Extension("PIL._imagingcms", ["src/_imagingcms.c"]),
    Extension("PIL._webp", ["src/_webp.c"]),
    Extension("PIL._imagingtk", ["src/_imagingtk.c", "src/Tk/tkImaging.c"]),
    Extension("PIL._imagingmath", ["src/_imagingmath.c"]),
    Extension("PIL._imagingmorph", ["src/_imagingmorph.c"]),
]

with open("README.md") as f:
    long_description = f.read()

try:
    setup(
        name=NAME,
        version=PILLOW_VERSION,
        description="Python Imaging Library (Fork)",
        long_description=long_description,
        long_description_content_type="text/markdown",
        license="HPND",
        author="Alex Clark (PIL Fork Author)",
        author_email="aclark@python-pillow.org",
        url="https://python-pillow.org",
        project_urls={
            "Documentation": "https://pillow.readthedocs.io",
            "Source": "https://github.com/python-pillow/Pillow",
            "Funding": "https://tidelift.com/subscription/pkg/pypi-pillow?"
            "utm_source=pypi-pillow&utm_medium=pypi",
        },
        classifiers=[
            "Development Status :: 6 - Mature",
            "License :: OSI Approved :: Historical Permission Notice and Disclaimer (HPND)",  # noqa: E501
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: Implementation :: PyPy",
            "Topic :: Multimedia :: Graphics",
            "Topic :: Multimedia :: Graphics :: Capture :: Digital Camera",
            "Topic :: Multimedia :: Graphics :: Capture :: Screen Capture",
            "Topic :: Multimedia :: Graphics :: Graphics Conversion",
            "Topic :: Multimedia :: Graphics :: Viewers",
        ],
        python_requires=">=3.6",
        cmdclass={"build_ext": pil_build_ext},
        ext_modules=ext_modules,
        include_package_data=True,
        packages=["PIL"],
        package_dir={"": "src"},
        keywords=["Imaging"],
        zip_safe=not (debug_build() or PLATFORM_MINGW),
    )
except RequiredDependencyException as err:
    msg = f"""

The headers or library files could not be found for {str(err)},
a required dependency when compiling Pillow from source.

Please see the install instructions at:
   https://pillow.readthedocs.io/en/latest/installation.html

"""
    sys.stderr.write(msg)
    raise RequiredDependencyException(msg)
except DependencyException as err:
    msg = f"""

The headers or library files could not be found for {str(err)},
which was requested by the option flag --enable-{str(err)}

"""
    sys.stderr.write(msg)
    raise DependencyException(msg)



#
# The Python Imaging Library.
# $Id$
#
# PDF (Acrobat) file handling
#
# History:
# 1996-07-16 fl   Created
# 1997-01-18 fl   Fixed header
# 2004-02-21 fl   Fixes for 1/L/CMYK images, etc.
# 2004-02-24 fl   Fixes for 1 and P images.
#
# Copyright (c) 1997-2004 by Secret Labs AB.  All rights reserved.
# Copyright (c) 1996-1997 by Fredrik Lundh.
#
# See the README file for information on usage and redistribution.
#

##
# Image plugin for PDF images (output only).
##

import io
import os
import time

from . import Image, ImageFile, ImageSequence, PdfParser, __version__

#
# --------------------------------------------------------------------

# object ids:
#  1. catalogue
#  2. pages
#  3. image
#  4. page
#  5. page contents


def _save_all(im, fp, filename):
    _save(im, fp, filename, save_all=True)


##
# (Internal) Image save plugin for the PDF format.


def _save(im, fp, filename, save_all=False):
    is_appending = im.encoderinfo.get("append", False)
    if is_appending:
        existing_pdf = PdfParser.PdfParser(f=fp, filename=filename, mode="r+b")
    else:
        existing_pdf = PdfParser.PdfParser(f=fp, filename=filename, mode="w+b")

    resolution = im.encoderinfo.get("resolution", 72.0)

    info = {
        "title": None
        if is_appending
        else os.path.splitext(os.path.basename(filename))[0],
        "author": None,
        "subject": None,
        "keywords": None,
        "creator": None,
        "producer": None,
        "creationDate": None if is_appending else time.gmtime(),
        "modDate": None if is_appending else time.gmtime(),
    }
    for k, default in info.items():
        v = im.encoderinfo.get(k) if k in im.encoderinfo else default
        if v:
            existing_pdf.info[k[0].upper() + k[1:]] = v

    #
    # make sure image data is available
    im.load()

    existing_pdf.start_writing()
    existing_pdf.write_header()
    existing_pdf.write_comment(f"created by Pillow {__version__} PDF driver")

    #
    # pages
    ims = [im]
    if save_all:
        append_images = im.encoderinfo.get("append_images", [])
        for append_im in append_images:
            append_im.encoderinfo = im.encoderinfo.copy()
            ims.append(append_im)
    numberOfPages = 0
    image_refs = []
    page_refs = []
    contents_refs = []
    for im in ims:
        im_numberOfPages = 1
        if save_all:
            try:
                im_numberOfPages = im.n_frames
            except AttributeError:
                # Image format does not have n_frames.
                # It is a single frame image
                pass
        numberOfPages += im_numberOfPages
        for i in range(im_numberOfPages):
            image_refs.append(existing_pdf.next_object_id(0))
            page_refs.append(existing_pdf.next_object_id(0))
            contents_refs.append(existing_pdf.next_object_id(0))
            existing_pdf.pages.append(page_refs[-1])

    #
    # catalog and list of pages
    existing_pdf.write_catalog()

    pageNumber = 0
    for imSequence in ims:
        im_pages = ImageSequence.Iterator(imSequence) if save_all else [imSequence]
        for im in im_pages:
            # FIXME: Should replace ASCIIHexDecode with RunLengthDecode
            # (packbits) or LZWDecode (tiff/lzw compression).  Note that
            # PDF 1.2 also supports Flatedecode (zip compression).

            bits = 8
            params = None
            decode = None

            if im.mode == "1":
                filter = "ASCIIHexDecode"
                colorspace = PdfParser.PdfName("DeviceGray")
                procset = "ImageB"  # grayscale
                bits = 1
            elif im.mode == "L":
                filter = "DCTDecode"
                # params = f"<< /Predictor 15 /Columns {width-2} >>"
                colorspace = PdfParser.PdfName("DeviceGray")
                procset = "ImageB"  # grayscale
            elif im.mode == "P":
                filter = "ASCIIHexDecode"
                palette = im.im.getpalette("RGB")
                colorspace = [
                    PdfParser.PdfName("Indexed"),
                    PdfParser.PdfName("DeviceRGB"),
                    255,
                    PdfParser.PdfBinary(palette),
                ]
                procset = "ImageI"  # indexed color
            elif im.mode == "RGB":
                filter = "DCTDecode"
                colorspace = PdfParser.PdfName("DeviceRGB")
                procset = "ImageC"  # color images
            elif im.mode == "CMYK":
                filter = "DCTDecode"
                colorspace = PdfParser.PdfName("DeviceCMYK")
                procset = "ImageC"  # color images
                decode = [1, 0, 1, 0, 1, 0, 1, 0]
            else:
                raise ValueError(f"cannot save mode {im.mode}")

            #
            # image

            op = io.BytesIO()

            if filter == "ASCIIHexDecode":
                if bits == 1:
                    # FIXME: the hex encoder doesn't support packed 1-bit
                    # images; do things the hard way...
                    data = im.tobytes("raw", "1")
                    im = Image.new("L", im.size)
                    im.putdata(data)
                ImageFile._save(im, op, [("hex", (0, 0) + im.size, 0, im.mode)])
            elif filter == "DCTDecode":
                Image.SAVE["JPEG"](im, op, filename)
            elif filter == "FlateDecode":
                ImageFile._save(im, op, [("zip", (0, 0) + im.size, 0, im.mode)])
            elif filter == "RunLengthDecode":
                ImageFile._save(im, op, [("packbits", (0, 0) + im.size, 0, im.mode)])
            else:
                raise ValueError(f"unsupported PDF filter ({filter})")

            #
            # Get image characteristics

            width, height = im.size

            existing_pdf.write_obj(
                image_refs[pageNumber],
                stream=op.getvalue(),
                Type=PdfParser.PdfName("XObject"),
                Subtype=PdfParser.PdfName("Image"),
                Width=width,  # * 72.0 / resolution,
                Height=height,  # * 72.0 / resolution,
                Filter=PdfParser.PdfName(filter),
                BitsPerComponent=bits,
                Decode=decode,
                DecodeParams=params,
                ColorSpace=colorspace,
            )

            #
            # page

            existing_pdf.write_page(
                page_refs[pageNumber],
                Resources=PdfParser.PdfDict(
                    ProcSet=[PdfParser.PdfName("PDF"), PdfParser.PdfName(procset)],
                    XObject=PdfParser.PdfDict(image=image_refs[pageNumber]),
                ),
                MediaBox=[
                    0,
                    0,
                    int(width * 72.0 / resolution),
                    int(height * 72.0 / resolution),
                ],
                Contents=contents_refs[pageNumber],
            )

            #
            # page contents

            page_contents = b"q %d 0 0 %d 0 0 cm /image Do Q\n" % (
                int(width * 72.0 / resolution),
                int(height * 72.0 / resolution),
            )

            existing_pdf.write_obj(contents_refs[pageNumber], stream=page_contents)

            pageNumber += 1

    #
    # trailer
    existing_pdf.write_xref_and_trailer()
    if hasattr(fp, "flush"):
        fp.flush()
    existing_pdf.close()


#
# --------------------------------------------------------------------


Image.register_save("PDF", _save)
Image.register_save_all("PDF", _save_all)

Image.register_extension("PDF", ".pdf")

Image.register_mime("PDF", "application/pdf")



import copy
import os
import re
import shutil
import sys
from io import BytesIO

import pytest
from packaging.version import parse as parse_version

from PIL import Image, ImageDraw, ImageFont, features

from .helper import (
    assert_image_equal,
    assert_image_equal_tofile,
    assert_image_similar,
    assert_image_similar_tofile,
    is_win32,
    skip_unless_feature,
    skip_unless_feature_version,
)

FONT_PATH = "Tests/fonts/FreeMono.ttf"
FONT_SIZE = 20

TEST_TEXT = "hey you\nyou are awesome\nthis looks awkward"


pytestmark = skip_unless_feature("freetype2")


class TestImageFont:
    LAYOUT_ENGINE = ImageFont.LAYOUT_BASIC

    # Freetype has different metrics depending on the version.
    # (and, other things, but first things first)
    METRICS = {
        (">=2.3", "<2.4"): {
            "multiline": 30,
            "textsize": 12,
            "getters": (13, 16),
            "mask": (107, 13),
            "multiline-anchor": 6,
            "getlength": (36, 27, 27, 33),
        },
        (">=2.7",): {
            "multiline": 6.2,
            "textsize": 2.5,
            "getters": (12, 16),
            "mask": (108, 13),
            "multiline-anchor": 4,
            "getlength": (36, 21, 24, 33),
        },
        "Default": {
            "multiline": 0.5,
            "textsize": 0.5,
            "getters": (12, 16),
            "mask": (108, 13),
            "multiline-anchor": 4,
            "getlength": (36, 24, 24, 33),
        },
    }

    @classmethod
    def setup_class(self):
        freetype = parse_version(features.version_module("freetype2"))

        self.metrics = self.METRICS["Default"]
        for conditions, metrics in self.METRICS.items():
            if not isinstance(conditions, tuple):
                continue

            for condition in conditions:
                version = parse_version(re.sub("[<=>]", "", condition))
                if (condition.startswith(">=") and freetype >= version) or (
                    condition.startswith("<") and freetype < version
                ):
                    # Condition was met
                    continue

                # Condition failed
                break
            else:
                # All conditions were met
                self.metrics = metrics

    def get_font(self):
        return ImageFont.truetype(
            FONT_PATH, FONT_SIZE, layout_engine=self.LAYOUT_ENGINE
        )

    def test_sanity(self):
        assert re.search(r"\d+\.\d+\.\d+$", features.version_module("freetype2"))

    def test_font_properties(self):
        ttf = self.get_font()
        assert ttf.path == FONT_PATH
        assert ttf.size == FONT_SIZE

        ttf_copy = ttf.font_variant()
        assert ttf_copy.path == FONT_PATH
        assert ttf_copy.size == FONT_SIZE

        ttf_copy = ttf.font_variant(size=FONT_SIZE + 1)
        assert ttf_copy.size == FONT_SIZE + 1

        second_font_path = "Tests/fonts/DejaVuSans.ttf"
        ttf_copy = ttf.font_variant(font=second_font_path)
        assert ttf_copy.path == second_font_path

    def test_font_with_name(self):
        self.get_font()
        self._render(FONT_PATH)

    def _font_as_bytes(self):
        with open(FONT_PATH, "rb") as f:
            font_bytes = BytesIO(f.read())
        return font_bytes

    def test_font_with_filelike(self):
        ImageFont.truetype(
            self._font_as_bytes(), FONT_SIZE, layout_engine=self.LAYOUT_ENGINE
        )
        self._render(self._font_as_bytes())
        # Usage note:  making two fonts from the same buffer fails.
        # shared_bytes = self._font_as_bytes()
        # self._render(shared_bytes)
        # with pytest.raises(Exception):
        #   _render(shared_bytes)

    def test_font_with_open_file(self):
        with open(FONT_PATH, "rb") as f:
            self._render(f)

    def test_non_ascii_path(self, tmp_path):
        tempfile = str(tmp_path / ("temp_" + chr(128) + ".ttf"))
        try:
            shutil.copy(FONT_PATH, tempfile)
        except UnicodeEncodeError:
            pytest.skip("Non-ASCII path could not be created")

        ImageFont.truetype(tempfile, FONT_SIZE)

    def test_unavailable_layout_engine(self):
        have_raqm = ImageFont.core.HAVE_RAQM
        ImageFont.core.HAVE_RAQM = False

        try:
            ttf = ImageFont.truetype(
                FONT_PATH, FONT_SIZE, layout_engine=ImageFont.LAYOUT_RAQM
            )
        finally:
            ImageFont.core.HAVE_RAQM = have_raqm

        assert ttf.layout_engine == ImageFont.LAYOUT_BASIC

    def _render(self, font):
        txt = "Hello World!"
        ttf = ImageFont.truetype(font, FONT_SIZE, layout_engine=self.LAYOUT_ENGINE)
        ttf.getsize(txt)

        img = Image.new("RGB", (256, 64), "white")
        d = ImageDraw.Draw(img)
        d.text((10, 10), txt, font=ttf, fill="black")

        return img

    def test_render_equal(self):
        img_path = self._render(FONT_PATH)
        with open(FONT_PATH, "rb") as f:
            font_filelike = BytesIO(f.read())
        img_filelike = self._render(font_filelike)

        assert_image_equal(img_path, img_filelike)

    def test_transparent_background(self):
        im = Image.new(mode="RGBA", size=(300, 100))
        draw = ImageDraw.Draw(im)
        ttf = self.get_font()

        txt = "Hello World!"
        draw.text((10, 10), txt, font=ttf)

        target = "Tests/images/transparent_background_text.png"
        with Image.open(target) as target_img:
            assert_image_similar(im, target_img, 4.09)

    def test_textsize_equal(self):
        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        ttf = self.get_font()

        txt = "Hello World!"
        size = draw.textsize(txt, ttf)
        draw.text((10, 10), txt, font=ttf)
        draw.rectangle((10, 10, 10 + size[0], 10 + size[1]))

        target = "Tests/images/rectangle_surrounding_text.png"
        with Image.open(target) as target_img:

            # Epsilon ~.5 fails with FreeType 2.7
            assert_image_similar(im, target_img, self.metrics["textsize"])

    @pytest.mark.parametrize(
        "text, mode, font, size, length_basic_index, length_raqm",
        (
            # basic test
            ("text", "L", "FreeMono.ttf", 15, 0, 36),
            ("text", "1", "FreeMono.ttf", 15, 0, 36),
            # issue 4177
            ("rrr", "L", "DejaVuSans.ttf", 18, 1, 22.21875),
            ("rrr", "1", "DejaVuSans.ttf", 18, 2, 22.21875),
            # test 'l' not including extra margin
            # using exact value 2047 / 64 for raqm, checked with debugger
            ("ill", "L", "OpenSansCondensed-LightItalic.ttf", 63, 3, 31.984375),
            ("ill", "1", "OpenSansCondensed-LightItalic.ttf", 63, 3, 31.984375),
        ),
    )
    def test_getlength(self, text, mode, font, size, length_basic_index, length_raqm):
        f = ImageFont.truetype(
            "Tests/fonts/" + font, size, layout_engine=self.LAYOUT_ENGINE
        )

        im = Image.new(mode, (1, 1), 0)
        d = ImageDraw.Draw(im)

        if self.LAYOUT_ENGINE == ImageFont.LAYOUT_BASIC:
            length = d.textlength(text, f)
            assert length == self.metrics["getlength"][length_basic_index]
        else:
            # disable kerning, kerning metrics changed
            length = d.textlength(text, f, features=["-kern"])
            assert length == length_raqm

    def test_render_multiline(self):
        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        ttf = self.get_font()
        line_spacing = draw.textsize("A", font=ttf)[1] + 4
        lines = TEST_TEXT.split("\n")
        y = 0
        for line in lines:
            draw.text((0, y), line, font=ttf)
            y += line_spacing

        target = "Tests/images/multiline_text.png"
        with Image.open(target) as target_img:

            # some versions of freetype have different horizontal spacing.
            # setting a tight epsilon, I'm showing the original test failure
            # at epsilon = ~38.
            assert_image_similar(im, target_img, self.metrics["multiline"])

    def test_render_multiline_text(self):
        ttf = self.get_font()

        # Test that text() correctly connects to multiline_text()
        # and that align defaults to left
        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.text((0, 0), TEST_TEXT, font=ttf)

        target = "Tests/images/multiline_text.png"
        with Image.open(target) as target_img:

            # Epsilon ~.5 fails with FreeType 2.7
            assert_image_similar(im, target_img, self.metrics["multiline"])

        # Test that text() can pass on additional arguments
        # to multiline_text()
        draw.text(
            (0, 0), TEST_TEXT, fill=None, font=ttf, anchor=None, spacing=4, align="left"
        )
        draw.text((0, 0), TEST_TEXT, None, ttf, None, 4, "left")

        # Test align center and right
        for align, ext in {"center": "_center", "right": "_right"}.items():
            im = Image.new(mode="RGB", size=(300, 100))
            draw = ImageDraw.Draw(im)
            draw.multiline_text((0, 0), TEST_TEXT, font=ttf, align=align)

            target = "Tests/images/multiline_text" + ext + ".png"
            with Image.open(target) as target_img:

                # Epsilon ~.5 fails with FreeType 2.7
                assert_image_similar(im, target_img, self.metrics["multiline"])

    def test_unknown_align(self):
        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        ttf = self.get_font()

        # Act/Assert
        with pytest.raises(ValueError):
            draw.multiline_text((0, 0), TEST_TEXT, font=ttf, align="unknown")

    def test_draw_align(self):
        im = Image.new("RGB", (300, 100), "white")
        draw = ImageDraw.Draw(im)
        ttf = self.get_font()
        line = "some text"
        draw.text((100, 40), line, (0, 0, 0), font=ttf, align="left")

    def test_multiline_size(self):
        ttf = self.get_font()
        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)

        # Test that textsize() correctly connects to multiline_textsize()
        assert draw.textsize(TEST_TEXT, font=ttf) == draw.multiline_textsize(
            TEST_TEXT, font=ttf
        )

        # Test that multiline_textsize corresponds to ImageFont.textsize()
        # for single line text
        assert ttf.getsize("A") == draw.multiline_textsize("A", font=ttf)

        # Test that textsize() can pass on additional arguments
        # to multiline_textsize()
        draw.textsize(TEST_TEXT, font=ttf, spacing=4)
        draw.textsize(TEST_TEXT, ttf, 4)

    def test_multiline_width(self):
        ttf = self.get_font()
        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)

        assert (
            draw.textsize("longest line", font=ttf)[0]
            == draw.multiline_textsize("longest line\nline", font=ttf)[0]
        )

    def test_multiline_spacing(self):
        ttf = self.get_font()

        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)
        draw.multiline_text((0, 0), TEST_TEXT, font=ttf, spacing=10)

        target = "Tests/images/multiline_text_spacing.png"
        with Image.open(target) as target_img:

            # Epsilon ~.5 fails with FreeType 2.7
            assert_image_similar(im, target_img, self.metrics["multiline"])

    def test_rotated_transposed_font(self):
        img_grey = Image.new("L", (100, 100))
        draw = ImageDraw.Draw(img_grey)
        word = "testing"
        font = self.get_font()

        orientation = Image.ROTATE_90
        transposed_font = ImageFont.TransposedFont(font, orientation=orientation)

        # Original font
        draw.font = font
        box_size_a = draw.textsize(word)

        # Rotated font
        draw.font = transposed_font
        box_size_b = draw.textsize(word)

        # Check (w,h) of box a is (h,w) of box b
        assert box_size_a[0] == box_size_b[1]
        assert box_size_a[1] == box_size_b[0]

    def test_unrotated_transposed_font(self):
        img_grey = Image.new("L", (100, 100))
        draw = ImageDraw.Draw(img_grey)
        word = "testing"
        font = self.get_font()

        orientation = None
        transposed_font = ImageFont.TransposedFont(font, orientation=orientation)

        # Original font
        draw.font = font
        box_size_a = draw.textsize(word)

        # Rotated font
        draw.font = transposed_font
        box_size_b = draw.textsize(word)

        # Check boxes a and b are same size
        assert box_size_a == box_size_b

    def test_rotated_transposed_font_get_mask(self):
        # Arrange
        text = "mask this"
        font = self.get_font()
        orientation = Image.ROTATE_90
        transposed_font = ImageFont.TransposedFont(font, orientation=orientation)

        # Act
        mask = transposed_font.getmask(text)

        # Assert
        assert mask.size == self.metrics["mask"][::-1]

    def test_unrotated_transposed_font_get_mask(self):
        # Arrange
        text = "mask this"
        font = self.get_font()
        orientation = None
        transposed_font = ImageFont.TransposedFont(font, orientation=orientation)

        # Act
        mask = transposed_font.getmask(text)

        # Assert
        assert mask.size == self.metrics["mask"]

    def test_free_type_font_get_name(self):
        # Arrange
        font = self.get_font()

        # Act
        name = font.getname()

        # Assert
        assert ("FreeMono", "Regular") == name

    def test_free_type_font_get_metrics(self):
        # Arrange
        font = self.get_font()

        # Act
        ascent, descent = font.getmetrics()

        # Assert
        assert isinstance(ascent, int)
        assert isinstance(descent, int)
        assert (ascent, descent) == (16, 4)  # too exact check?

    def test_free_type_font_get_offset(self):
        # Arrange
        font = self.get_font()
        text = "offset this"

        # Act
        offset = font.getoffset(text)

        # Assert
        assert offset == (0, 3)

    def test_free_type_font_get_mask(self):
        # Arrange
        font = self.get_font()
        text = "mask this"

        # Act
        mask = font.getmask(text)

        # Assert
        assert mask.size == self.metrics["mask"]

    def test_load_path_not_found(self):
        # Arrange
        filename = "somefilenamethatdoesntexist.ttf"

        # Act/Assert
        with pytest.raises(OSError):
            ImageFont.load_path(filename)
        with pytest.raises(OSError):
            ImageFont.truetype(filename)

    def test_load_non_font_bytes(self):
        with open("Tests/images/hopper.jpg", "rb") as f:
            with pytest.raises(OSError):
                ImageFont.truetype(f)

    def test_default_font(self):
        # Arrange
        txt = 'This is a "better than nothing" default font.'
        im = Image.new(mode="RGB", size=(300, 100))
        draw = ImageDraw.Draw(im)

        target = "Tests/images/default_font.png"
        with Image.open(target) as target_img:

            # Act
            default_font = ImageFont.load_default()
            draw.text((10, 10), txt, font=default_font)

            # Assert
            assert_image_equal(im, target_img)

    def test_getsize_empty(self):
        # issue #2614
        font = self.get_font()
        # should not crash.
        assert (0, 0) == font.getsize("")

    def test_render_empty(self):
        # issue 2666
        font = self.get_font()
        im = Image.new(mode="RGB", size=(300, 100))
        target = im.copy()
        draw = ImageDraw.Draw(im)
        # should not crash here.
        draw.text((10, 10), "", font=font)
        assert_image_equal(im, target)

    def test_unicode_pilfont(self):
        # should not segfault, should return UnicodeDecodeError
        # issue #2826
        font = ImageFont.load_default()
        with pytest.raises(UnicodeEncodeError):
            font.getsize("’")

    def test_unicode_extended(self):
        # issue #3777
        text = "A\u278A\U0001F12B"
        target = "Tests/images/unicode_extended.png"

        ttf = ImageFont.truetype(
            "Tests/fonts/NotoSansSymbols-Regular.ttf",
            FONT_SIZE,
            layout_engine=self.LAYOUT_ENGINE,
        )
        img = Image.new("RGB", (100, 60))
        d = ImageDraw.Draw(img)
        d.text((10, 10), text, font=ttf)

        # fails with 14.7
        assert_image_similar_tofile(img, target, 6.2)

    def _test_fake_loading_font(self, monkeypatch, path_to_fake, fontname):
        # Make a copy of FreeTypeFont so we can patch the original
        free_type_font = copy.deepcopy(ImageFont.FreeTypeFont)
        with monkeypatch.context() as m:
            m.setattr(ImageFont, "_FreeTypeFont", free_type_font, raising=False)

            def loadable_font(filepath, size, index, encoding, *args, **kwargs):
                if filepath == path_to_fake:
                    return ImageFont._FreeTypeFont(
                        FONT_PATH, size, index, encoding, *args, **kwargs
                    )
                return ImageFont._FreeTypeFont(
                    filepath, size, index, encoding, *args, **kwargs
                )

            m.setattr(ImageFont, "FreeTypeFont", loadable_font)
            font = ImageFont.truetype(fontname)
            # Make sure it's loaded
            name = font.getname()
            assert ("FreeMono", "Regular") == name

    @pytest.mark.skipif(is_win32(), reason="requires Unix or macOS")
    def test_find_linux_font(self, monkeypatch):
        # A lot of mocking here - this is more for hitting code and
        # catching syntax like errors
        font_directory = "/usr/local/share/fonts"
        monkeypatch.setattr(sys, "platform", "linux")
        monkeypatch.setenv("XDG_DATA_DIRS", "/usr/share/:/usr/local/share/")

        def fake_walker(path):
            if path == font_directory:
                return [
                    (
                        path,
                        [],
                        ["Arial.ttf", "Single.otf", "Duplicate.otf", "Duplicate.ttf"],
                    )
                ]
            return [(path, [], ["some_random_font.ttf"])]

        monkeypatch.setattr(os, "walk", fake_walker)
        # Test that the font loads both with and without the
        # extension
        self._test_fake_loading_font(
            monkeypatch, font_directory + "/Arial.ttf", "Arial.ttf"
        )
        self._test_fake_loading_font(
            monkeypatch, font_directory + "/Arial.ttf", "Arial"
        )

        # Test that non-ttf fonts can be found without the
        # extension
        self._test_fake_loading_font(
            monkeypatch, font_directory + "/Single.otf", "Single"
        )

        # Test that ttf fonts are preferred if the extension is
        # not specified
        self._test_fake_loading_font(
            monkeypatch, font_directory + "/Duplicate.ttf", "Duplicate"
        )

    @pytest.mark.skipif(is_win32(), reason="requires Unix or macOS")
    def test_find_macos_font(self, monkeypatch):
        # Like the linux test, more cover hitting code rather than testing
        # correctness.
        font_directory = "/System/Library/Fonts"
        monkeypatch.setattr(sys, "platform", "darwin")

        def fake_walker(path):
            if path == font_directory:
                return [
                    (
                        path,
                        [],
                        ["Arial.ttf", "Single.otf", "Duplicate.otf", "Duplicate.ttf"],
                    )
                ]
            return [(path, [], ["some_random_font.ttf"])]

        monkeypatch.setattr(os, "walk", fake_walker)
        self._test_fake_loading_font(
            monkeypatch, font_directory + "/Arial.ttf", "Arial.ttf"
        )
        self._test_fake_loading_font(
            monkeypatch, font_directory + "/Arial.ttf", "Arial"
        )
        self._test_fake_loading_font(
            monkeypatch, font_directory + "/Single.otf", "Single"
        )
        self._test_fake_loading_font(
            monkeypatch, font_directory + "/Duplicate.ttf", "Duplicate"
        )

    def test_imagefont_getters(self):
        # Arrange
        t = self.get_font()

        # Act / Assert
        assert t.getmetrics() == (16, 4)
        assert t.font.ascent == 16
        assert t.font.descent == 4
        assert t.font.height == 20
        assert t.font.x_ppem == 20
        assert t.font.y_ppem == 20
        assert t.font.glyphs == 4177
        assert t.getsize("A") == (12, 16)
        assert t.getsize("AB") == (24, 16)
        assert t.getsize("M") == self.metrics["getters"]
        assert t.getsize("y") == (12, 20)
        assert t.getsize("a") == (12, 16)
        assert t.getsize_multiline("A") == (12, 16)
        assert t.getsize_multiline("AB") == (24, 16)
        assert t.getsize_multiline("a") == (12, 16)
        assert t.getsize_multiline("ABC\n") == (36, 36)
        assert t.getsize_multiline("ABC\nA") == (36, 36)
        assert t.getsize_multiline("ABC\nAaaa") == (48, 36)

    def test_getsize_stroke(self):
        # Arrange
        t = self.get_font()

        # Act / Assert
        for stroke_width in [0, 2]:
            assert t.getsize("A", stroke_width=stroke_width) == (
                12 + stroke_width * 2,
                16 + stroke_width * 2,
            )
            assert t.getsize_multiline("ABC\nAaaa", stroke_width=stroke_width) == (
                48 + stroke_width * 2,
                36 + stroke_width * 4,
            )

    def test_complex_font_settings(self):
        # Arrange
        t = self.get_font()
        # Act / Assert
        if t.layout_engine == ImageFont.LAYOUT_BASIC:
            with pytest.raises(KeyError):
                t.getmask("абвг", direction="rtl")
            with pytest.raises(KeyError):
                t.getmask("абвг", features=["-kern"])
            with pytest.raises(KeyError):
                t.getmask("абвг", language="sr")

    def test_variation_get(self):
        font = self.get_font()

        freetype = parse_version(features.version_module("freetype2"))
        if freetype < parse_version("2.9.1"):
            with pytest.raises(NotImplementedError):
                font.get_variation_names()
            with pytest.raises(NotImplementedError):
                font.get_variation_axes()
            return

        with pytest.raises(OSError):
            font.get_variation_names()
        with pytest.raises(OSError):
            font.get_variation_axes()

        font = ImageFont.truetype("Tests/fonts/AdobeVFPrototype.ttf")
        assert font.get_variation_names(), [
            b"ExtraLight",
            b"Light",
            b"Regular",
            b"Semibold",
            b"Bold",
            b"Black",
            b"Black Medium Contrast",
            b"Black High Contrast",
            b"Default",
        ]
        assert font.get_variation_axes() == [
            {"name": b"Weight", "minimum": 200, "maximum": 900, "default": 389},
            {"name": b"Contrast", "minimum": 0, "maximum": 100, "default": 0},
        ]

        font = ImageFont.truetype("Tests/fonts/TINY5x3GX.ttf")
        assert font.get_variation_names() == [
            b"20",
            b"40",
            b"60",
            b"80",
            b"100",
            b"120",
            b"140",
            b"160",
            b"180",
            b"200",
            b"220",
            b"240",
            b"260",
            b"280",
            b"300",
            b"Regular",
        ]
        assert font.get_variation_axes() == [
            {"name": b"Size", "minimum": 0, "maximum": 300, "default": 0}
        ]

    def _check_text(self, font, path, epsilon):
        im = Image.new("RGB", (100, 75), "white")
        d = ImageDraw.Draw(im)
        d.text((10, 10), "Text", font=font, fill="black")

        try:
            with Image.open(path) as expected:
                assert_image_similar(im, expected, epsilon)
        except AssertionError:
            if "_adobe" in path:
                path = path.replace("_adobe", "_adobe_older_harfbuzz")
                with Image.open(path) as expected:
                    assert_image_similar(im, expected, epsilon)
            else:
                raise

    def test_variation_set_by_name(self):
        font = self.get_font()

        freetype = parse_version(features.version_module("freetype2"))
        if freetype < parse_version("2.9.1"):
            with pytest.raises(NotImplementedError):
                font.set_variation_by_name("Bold")
            return

        with pytest.raises(OSError):
            font.set_variation_by_name("Bold")

        font = ImageFont.truetype("Tests/fonts/AdobeVFPrototype.ttf", 36)
        self._check_text(font, "Tests/images/variation_adobe.png", 11)
        for name in ["Bold", b"Bold"]:
            font.set_variation_by_name(name)
        self._check_text(font, "Tests/images/variation_adobe_name.png", 11)

        font = ImageFont.truetype("Tests/fonts/TINY5x3GX.ttf", 36)
        self._check_text(font, "Tests/images/variation_tiny.png", 40)
        for name in ["200", b"200"]:
            font.set_variation_by_name(name)
        self._check_text(font, "Tests/images/variation_tiny_name.png", 40)

    def test_variation_set_by_axes(self):
        font = self.get_font()

        freetype = parse_version(features.version_module("freetype2"))
        if freetype < parse_version("2.9.1"):
            with pytest.raises(NotImplementedError):
                font.set_variation_by_axes([100])
            return

        with pytest.raises(OSError):
            font.set_variation_by_axes([500, 50])

        font = ImageFont.truetype("Tests/fonts/AdobeVFPrototype.ttf", 36)
        font.set_variation_by_axes([500, 50])
        self._check_text(font, "Tests/images/variation_adobe_axes.png", 11.05)

        font = ImageFont.truetype("Tests/fonts/TINY5x3GX.ttf", 36)
        font.set_variation_by_axes([100])
        self._check_text(font, "Tests/images/variation_tiny_axes.png", 32.5)

    @pytest.mark.parametrize(
        "anchor, left, left_old, top",
        (
            # test horizontal anchors
            ("ls", 0, 0, -36),
            ("ms", -64, -65, -36),
            ("rs", -128, -129, -36),
            # test vertical anchors
            ("ma", -64, -65, 16),
            ("mt", -64, -65, 0),
            ("mm", -64, -65, -17),
            ("mb", -64, -65, -44),
            ("md", -64, -65, -51),
        ),
        ids=("ls", "ms", "rs", "ma", "mt", "mm", "mb", "md"),
    )
    def test_anchor(self, anchor, left, left_old, top):
        name, text = "quick", "Quick"
        path = f"Tests/images/test_anchor_{name}_{anchor}.png"

        freetype = parse_version(features.version_module("freetype2"))
        if freetype < parse_version("2.4"):
            width, height = (129, 44)
            left = left_old
        elif self.LAYOUT_ENGINE == ImageFont.LAYOUT_RAQM:
            width, height = (129, 44)
        else:
            width, height = (128, 44)

        bbox_expected = (left, top, left + width, top + height)

        f = ImageFont.truetype(
            "Tests/fonts/NotoSans-Regular.ttf", 48, layout_engine=self.LAYOUT_ENGINE
        )

        im = Image.new("RGB", (200, 200), "white")
        d = ImageDraw.Draw(im)
        d.line(((0, 100), (200, 100)), "gray")
        d.line(((100, 0), (100, 200)), "gray")
        d.text((100, 100), text, fill="black", anchor=anchor, font=f)

        assert d.textbbox((0, 0), text, f, anchor=anchor) == bbox_expected

        with Image.open(path) as expected:
            assert_image_similar(im, expected, 7)

    @pytest.mark.parametrize(
        "anchor, align",
        (
            # test horizontal anchors
            ("lm", "left"),
            ("lm", "center"),
            ("lm", "right"),
            ("mm", "left"),
            ("mm", "center"),
            ("mm", "right"),
            ("rm", "left"),
            ("rm", "center"),
            ("rm", "right"),
            # test vertical anchors
            ("ma", "center"),
            # ("mm", "center"),  # duplicate
            ("md", "center"),
        ),
    )
    def test_anchor_multiline(self, anchor, align):
        target = f"Tests/images/test_anchor_multiline_{anchor}_{align}.png"
        text = "a\nlong\ntext sample"

        f = ImageFont.truetype(
            "Tests/fonts/NotoSans-Regular.ttf", 48, layout_engine=self.LAYOUT_ENGINE
        )

        # test render
        im = Image.new("RGB", (600, 400), "white")
        d = ImageDraw.Draw(im)
        d.line(((0, 200), (600, 200)), "gray")
        d.line(((300, 0), (300, 400)), "gray")
        d.multiline_text(
            (300, 200), text, fill="black", anchor=anchor, font=f, align=align
        )

        with Image.open(target) as expected:
            assert_image_similar(im, expected, self.metrics["multiline-anchor"])

    def test_anchor_invalid(self):
        font = self.get_font()
        im = Image.new("RGB", (100, 100), "white")
        d = ImageDraw.Draw(im)
        d.font = font

        for anchor in ["", "l", "a", "lax", "sa", "xa", "lx"]:
            pytest.raises(ValueError, lambda: font.getmask2("hello", anchor=anchor))
            pytest.raises(ValueError, lambda: font.getbbox("hello", anchor=anchor))
            pytest.raises(ValueError, lambda: d.text((0, 0), "hello", anchor=anchor))
            pytest.raises(
                ValueError, lambda: d.textbbox((0, 0), "hello", anchor=anchor)
            )
            pytest.raises(
                ValueError, lambda: d.multiline_text((0, 0), "foo\nbar", anchor=anchor)
            )
            pytest.raises(
                ValueError,
                lambda: d.multiline_textbbox((0, 0), "foo\nbar", anchor=anchor),
            )
        for anchor in ["lt", "lb"]:
            pytest.raises(
                ValueError, lambda: d.multiline_text((0, 0), "foo\nbar", anchor=anchor)
            )
            pytest.raises(
                ValueError,
                lambda: d.multiline_textbbox((0, 0), "foo\nbar", anchor=anchor),
            )

    @skip_unless_feature("freetype2")
    @pytest.mark.parametrize("bpp", (1, 2, 4, 8))
    def test_bitmap_font(self, bpp):
        text = "Bitmap Font"
        layout_name = ["basic", "raqm"][self.LAYOUT_ENGINE]
        target = f"Tests/images/bitmap_font_{bpp}_{layout_name}.png"
        font = ImageFont.truetype(
            f"Tests/fonts/DejaVuSans-24-{bpp}-stripped.ttf",
            24,
            layout_engine=self.LAYOUT_ENGINE,
        )

        im = Image.new("RGB", (160, 35), "white")
        draw = ImageDraw.Draw(im)
        draw.text((2, 2), text, "black", font)

        assert_image_equal_tofile(im, target)

    def test_standard_embedded_color(self):
        txt = "Hello World!"
        ttf = ImageFont.truetype(FONT_PATH, 40, layout_engine=self.LAYOUT_ENGINE)
        ttf.getsize(txt)

        im = Image.new("RGB", (300, 64), "white")
        d = ImageDraw.Draw(im)
        d.text((10, 10), txt, font=ttf, fill="#fa6", embedded_color=True)

        with Image.open("Tests/images/standard_embedded.png") as expected:
            assert_image_similar(im, expected, max(self.metrics["multiline"], 3))

    @skip_unless_feature_version("freetype2", "2.5.0")
    def test_cbdt(self):
        try:
            font = ImageFont.truetype(
                "Tests/fonts/NotoColorEmoji.ttf",
                size=109,
                layout_engine=self.LAYOUT_ENGINE,
            )

            im = Image.new("RGB", (150, 150), "white")
            d = ImageDraw.Draw(im)

            d.text((10, 10), "\U0001f469", embedded_color=True, font=font)

            with Image.open("Tests/images/cbdt_notocoloremoji.png") as expected:
                assert_image_similar(im, expected, self.metrics["multiline"])
        except IOError as e:
            assert str(e) in ("unimplemented feature", "unknown file format")
            pytest.skip("freetype compiled without libpng or unsupported")

    @skip_unless_feature_version("freetype2", "2.5.0")
    def test_cbdt_mask(self):
        try:
            font = ImageFont.truetype(
                "Tests/fonts/NotoColorEmoji.ttf",
                size=109,
                layout_engine=self.LAYOUT_ENGINE,
            )

            im = Image.new("RGB", (150, 150), "white")
            d = ImageDraw.Draw(im)

            d.text((10, 10), "\U0001f469", "black", font=font)

            with Image.open("Tests/images/cbdt_notocoloremoji_mask.png") as expected:
                assert_image_similar(im, expected, self.metrics["multiline"])
        except IOError as e:
            assert str(e) in ("unimplemented feature", "unknown file format")
            pytest.skip("freetype compiled without libpng or unsupported")

    @skip_unless_feature_version("freetype2", "2.10.0")
    def test_colr(self):
        font = ImageFont.truetype(
            "Tests/fonts/BungeeColor-Regular_colr_Windows.ttf",
            size=64,
            layout_engine=self.LAYOUT_ENGINE,
        )

        im = Image.new("RGB", (300, 75), "white")
        d = ImageDraw.Draw(im)

        d.text((15, 5), "Bungee", embedded_color=True, font=font)

        with Image.open("Tests/images/colr_bungee.png") as expected:
            assert_image_similar(im, expected, 21)

    @skip_unless_feature_version("freetype2", "2.10.0")
    def test_colr_mask(self):
        font = ImageFont.truetype(
            "Tests/fonts/BungeeColor-Regular_colr_Windows.ttf",
            size=64,
            layout_engine=self.LAYOUT_ENGINE,
        )

        im = Image.new("RGB", (300, 75), "white")
        d = ImageDraw.Draw(im)

        d.text((15, 5), "Bungee", "black", font=font)

        with Image.open("Tests/images/colr_bungee_mask.png") as expected:
            assert_image_similar(im, expected, 22)


@skip_unless_feature("raqm")
class TestImageFont_RaqmLayout(TestImageFont):
    LAYOUT_ENGINE = ImageFont.LAYOUT_RAQM


@skip_unless_feature_version("freetype2", "2.4", "Different metrics")
def test_render_mono_size():
    # issue 4177

    im = Image.new("P", (100, 30), "white")
    draw = ImageDraw.Draw(im)
    ttf = ImageFont.truetype(
        "Tests/fonts/DejaVuSans.ttf", 18, layout_engine=ImageFont.LAYOUT_BASIC
    )

    draw.text((10, 10), "r" * 10, "black", ttf)
    assert_image_equal_tofile(im, "Tests/images/text_mono.gif")



import pytest

from PIL import Image, PsdImagePlugin

from .helper import assert_image_similar, hopper, is_pypy

test_file = "Tests/images/hopper.psd"


def test_sanity():
    with Image.open(test_file) as im:
        im.load()
        assert im.mode == "RGB"
        assert im.size == (128, 128)
        assert im.format == "PSD"
        assert im.get_format_mimetype() == "image/vnd.adobe.photoshop"

        im2 = hopper()
        assert_image_similar(im, im2, 4.8)


@pytest.mark.skipif(is_pypy(), reason="Requires CPython")
def test_unclosed_file():
    def open():
        im = Image.open(test_file)
        im.load()

    pytest.warns(ResourceWarning, open)


def test_closed_file():
    def open():
        im = Image.open(test_file)
        im.load()
        im.close()

    pytest.warns(None, open)


def test_context_manager():
    def open():
        with Image.open(test_file) as im:
            im.load()

    pytest.warns(None, open)


def test_invalid_file():
    invalid_file = "Tests/images/flower.jpg"

    with pytest.raises(SyntaxError):
        PsdImagePlugin.PsdImageFile(invalid_file)


def test_n_frames():
    with Image.open("Tests/images/hopper_merged.psd") as im:
        assert im.n_frames == 1
        assert not im.is_animated

    with Image.open(test_file) as im:
        assert im.n_frames == 2
        assert im.is_animated


def test_eoferror():
    with Image.open(test_file) as im:
        # PSD seek index starts at 1 rather than 0
        n_frames = im.n_frames + 1

        # Test seeking past the last frame
        with pytest.raises(EOFError):
            im.seek(n_frames)
        assert im.tell() < n_frames

        # Test that seeking to the last frame does not raise an error
        im.seek(n_frames - 1)


def test_seek_tell():
    with Image.open(test_file) as im:

        layer_number = im.tell()
        assert layer_number == 1

        with pytest.raises(EOFError):
            im.seek(0)

        im.seek(1)
        layer_number = im.tell()
        assert layer_number == 1

        im.seek(2)
        layer_number = im.tell()
        assert layer_number == 2


def test_seek_eoferror():
    with Image.open(test_file) as im:

        with pytest.raises(EOFError):
            im.seek(-1)


def test_open_after_exclusive_load():
    with Image.open(test_file) as im:
        im.load()
        im.seek(im.tell() + 1)
        im.load()


def test_icc_profile():
    with Image.open(test_file) as im:
        assert "icc_profile" in im.info

        icc_profile = im.info["icc_profile"]
        assert len(icc_profile) == 3144


def test_no_icc_profile():
    with Image.open("Tests/images/hopper_merged.psd") as im:
        assert "icc_profile" not in im.info


def test_combined_larger_than_size():
    # The 'combined' sizes of the individual parts is larger than the
    # declared 'size' of the extra data field, resulting in a backwards seek.

    # If we instead take the 'size' of the extra data field as the source of truth,
    # then the seek can't be negative
    with pytest.raises(OSError):
        Image.open("Tests/images/combined_larger_than_size.psd")
